/*-----------------------------------------------------------------*/
/*! 
  \file omlfileinputstream.cxx
  \brief Open Math Input Stream with file descriptor (int)  implementation.
  \author  M. Gastineau 
           Astronomie et Systemes Dynamiques, IMCCE, CNRS, Observatoire de Paris. 

   Copyright, 2008,  M. Gastineau, IMCCE-CNRS
   email of the author : gastineau@imcce.fr
*/
/*-----------------------------------------------------------------*/

/*-----------------------------------------------------------------*/
/* License  of this file :
This file is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute this file under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/*-----------------------------------------------------------------*/


#if HAVE_CONFIG_H
#include "scscpconfig.h"
#endif /*HAVE_CONFIG_H*/
#include <stdio.h>
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <libxml/parser.h>
#include <libxml/tree.h>
#if HAVE_STRING_H
#include <string.h>
#endif
#include <errno.h>
#include <ctype.h>
#include "scscptypes.h"
#include "scscpinternal.h"
#include "omlfileinputstream.h"
#include "scscpdebug.h"
#include "scscp.h"

/*-----------------------------------------------------------------*/
/* structure locale */
/*-----------------------------------------------------------------*/
struct t_controlpipe {
 t_controlpipe(OML_Fileio&  pfile, bool pbfind, char *plastbuf, bool plastbufvalid)
    : file(pfile), mathfind(pbfind), lastbufvalid(plastbufvalid), status(0)
 {
  strcpy(lastbuf, plastbuf);
 };
 
 OML_Fileio&  file;      /*!< flux de type fichier */
 bool mathfind ;  /*!< =true => </math> trouve */
 char lastbuf[100]; /*!< dernier buffer lu */
 bool lastbufvalid; /*!< lastbuf est valide */
 SCSCP_status status; /*!< status of the connection */
} ;


/***************************************************************************/
/* macros locales aux fichiers                                                      
                           */
/***************************************************************************/
//! buffer for len for scscp protocol
#define MAXLENBUFFER 4094

/*-----------------------------------------------------------------*/
/*! read the end of line
  @param fd (inout) file descriptor
  @param c (out) read buffer (at least 4094 bytes)
  @param status (inout) status
*/
/*-----------------------------------------------------------------*/
static bool ReadLineBuffer(OML_Fileio& fd, char *c, SCSCP_status* status)
{
    const char *symbolquit="<?scscp quit";
    bool bres = true;
    int j=-1;
    SCSCP_getdebugger()<<"ReadLineBuffer("<<fd.GetFile()<<",) - entree\n";
    
    do
    {
     j++;
     int ival = fd.read(c+j, sizeof(char));
     bres=(ival>0);
    } while (bres && c[j]!='\n' && j<=MAXLENBUFFER-1);
    SCSCP_getdebugger()<<"ReadLineBuffer- a lu '"<<c<<"'\n";
    c[j]='\0';
    
    if (strncmp(symbolquit, c,strlen(symbolquit))==0) 
    {
     SCSCP_getdebugger()<<"'<?scscp quit' found\n";
     SCSCP_status_seterror(status, SCSCP_STATUS_RECVQUIT);
    }
    SCSCP_getdebugger()<<"ReadLineBuffer(,"<<SCSCP_status_is(status)<<")  - sortie et retourne "<<bres<<"\n";
    return bres;
}

/*-----------------------------------------------------------------------*/
/*! routine de lecture appelle par xmlReadIO.
 Doit etre compatible avec xmlInputReadCallback
 retourne la taille du buffer valide
 returne -1 lorsque </math> a deja ete trouve au precedent appel
 @param ctx (inout) contexte du flux
 @param buff (inout) buffer
 @param len (in) taille du buffer
*/
/*-----------------------------------------------------------------------*/
static int lecture(t_controlpipe* ctx, char* buff, int len)
{
    OML_Fileio&  file=ctx->file;
	int i=0;
	char *bufcmp;
	const char *symbol = "<?scscp";
	int lensymbol = strlen(symbol);
	int eof;
	
    if (!ctx->mathfind) 
    {
    do
    {
     errno=0;
     eof=(file.read(buff+i, sizeof(char))<=-1);
     /*printf("c=%c '%.*s'\n", buff[i], 7,(i>=7)?buff+i-7:"vide");*/
     SCSCP_getdebugger()<<buff[i];
     i++;
     if (i<lensymbol && ctx->lastbufvalid)
     {
      memmove(ctx->lastbuf, ctx->lastbuf+1, (lensymbol-1)*sizeof(char));
      ctx->lastbuf[lensymbol-1] = buff[i-1];
     }
     else bufcmp = buff+i-lensymbol;
     
     if (i>=lensymbol) 
     {
        if (strncmp(bufcmp, symbol, lensymbol)==0)
        {
           ctx->mathfind=true;
           i-=lensymbol;
           //read the end of the line to cancel or not
           char c[MAXLENBUFFER];
           bool bres = ReadLineBuffer(file, c, &ctx->status);
           if (!bres || strcmp(c," cancel ?>")==0) 
           {
              SCSCP_getdebugger()<<"cancel\n";
              SCSCP_status_seterror(&ctx->status, SCSCP_STATUS_RECVCANCEL);
              i = -1;
           }
           if (strcmp(c," end ?>")==0) 
           {
              SCSCP_getdebugger()<<"end received\n";
           }
           break;
        }
     }
     
    } while (!eof && i<len );
 
     /* balise non trouvee : buffer plein */
     if (!ctx->mathfind)
     {
      if (i>=lensymbol) 
      {
       memcpy(ctx->lastbuf, buff+i-lensymbol, lensymbol*sizeof(char));
       ctx->lastbufvalid = true;
      }
     }
    }
    else i=-1;
    
   return i;
}



/*-----------------------------------------------------------------*/
/*! read the xml document. return true on success
*/
/*-----------------------------------------------------------------*/
bool OML_Fileinputstream::ReadXmlDoc()
{
  bool res;
  
  SCSCP_getdebugger()<<"\nOML_Fileinputstream::ReadXmlDoc() - entree\n";
  
  LIBXML_TEST_VERSION
  t_controlpipe ctx(GetFileio(), false, "                          ", false);
  xmlDocPtr doc = xmlReadIO((xmlInputReadCallback)lecture,NULL, &ctx, NULL, NULL, /*XML_PARSE_NOERROR|XML_PARSE_NOWARNING*/ XML_PARSE_NOBLANKS);

  res = (doc!=NULL);
  if (res)
  {
   SetXmlDoc(doc);
  }
  
  SCSCP_getdebugger()<<"\nOML_Fileinputstream::ReadXmlDoc() - sortie et retourne "<<res<<"\n";
 return res;
}

/*-----------------------------------------------------------------*/
/*! read a non-blank line finished by '\n' 
  and return true on success
*/
/*-----------------------------------------------------------------*/
bool OML_Fileinputstream::ReadLine(SCSCP_status* status)
{
    char c[MAXLENBUFFER];
    SCSCP_getdebugger()<<"OML_Fileinputstream::ReadLine() - entree\n";
    
    bool bres;
    do
    {
     bres = ReadLineBuffer(GetFileio(), c,status);
     //check that c is non blank
     int j=0;
     while (c[j]!='\0' && isspace(c[j])!=0) j++;
     if (c[j]!='\0') break;
     
    } while (bres);
    if (!bres) SetError();
    
    SCSCP_getdebugger()<<"OML_Fileinputstream::ReadLine("<<SCSCP_status_is(status)<<") - sortie et retourne "<<bres<<"\n";
    return bres;
}
