/*-----------------------------------------------------------------*/
/*! 
  \file omlinputstream.cxx
  \brief Open Math Input Stream   implementation.
  \author  M. Gastineau 
           Astronomie et Systemes Dynamiques, IMCCE, CNRS, Observatoire de Paris. 

   Copyright, 2008,  M. Gastineau, IMCCE-CNRS
   email of the author : gastineau@imcce.fr
*/
/*-----------------------------------------------------------------*/

/*-----------------------------------------------------------------*/
/* License  of this file :
This file is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute this file under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/*-----------------------------------------------------------------*/


#if HAVE_CONFIG_H
#include "scscpconfig.h"
#endif /*HAVE_CONFIG_H*/
#include <stdio.h>
#if HAVE_STRING_H
#include <string.h>
#endif
#include <libxml/parser.h>
#include <libxml/tree.h>
#include "scscptypes.h"
#include "scscpinternal.h"
#include "omlinputstream.h"
#include "omloutputstream.h"
#include "scscpdebug.h"
#include "omloms.h"
#include "omlomi.h"
#include "omlomstr.h"
#define __SCSCP_WITHIN_SCSCP
#include "scscp.h"

/***************************************************************************/
/* macros locales aux fichiers */
/***************************************************************************/
/*! conversion de xmlchar en char */
#define xmlChar2char(x) ((const char*)(x))

/*-----------------------------------------------------------------*/
/* variables locales */
/*-----------------------------------------------------------------*/
//! table of openmath object
const struct 
{ 
	OML_inputstream::OMLtype type; 
	const char *name;
} 
s_tableindicename[]=
{
 { OML_inputstream::eOMLOMI, "OMI" },
 { OML_inputstream::eOMLOMS, "OMS" },
 { OML_inputstream::eOMLOMA, "OMA" },
 { OML_inputstream::eOMLOMV, "OMV" },
 { OML_inputstream::eOMLOMF, "OMF" },
 { OML_inputstream::eOMLOMSTR, "OMSTR" },
 { OML_inputstream::eOMLInvalid, "" }
};

/*----------------------------------------------------------------------------------------------*/
/*! return the next node after the current xml node.
  return NULL if curnode is the last node.
  @param curnode (in) current valid node. mustn't be NULL 
*/
/*----------------------------------------------------------------------------------------------*/
SCSCP_xmlnodeptr SCSCP_xmlnode_getnext(SCSCP_xmlnodeptr curnode)
{
 return curnode->next;
}

/*----------------------------------------------------------------------------------------------*/
/*! return the name after the current xml node.
  @param curnode (in) current valid node. mustn't be NULL 
*/
/*----------------------------------------------------------------------------------------------*/
const char* SCSCP_xmlnode_getname(SCSCP_xmlnodeptr curnode)
{ 
 return xmlChar2char(curnode->name);
}


/*----------------------------------------------------------------------------------------------*/
/*! return the first child node the current xml node.
  return NULL if no child present
  @param curnode (in) current valid node. mustn't be NULL 
*/
/*----------------------------------------------------------------------------------------------*/
SCSCP_xmlnodeptr SCSCP_xmlnode_getchild(SCSCP_xmlnodeptr curnode)
{ 
 return curnode->children;
}

/*----------------------------------------------------------------------------------------------*/
/*! read the content of the current node in the current xml node
  @param curnode (in) current valid node. mustn't be NULL 
*/
/*----------------------------------------------------------------------------------------------*/
const char * SCSCP_xmlnode_getcontent(SCSCP_xmlnodeptr curnode)
{ 
 return xmlChar2char(XML_GET_CONTENT(curnode));
}

/*----------------------------------------------------------------------------------------------*/
/*! read the first attribute of the current node
  return NULL if no attribute present
  @param curnode (in) current valid node. mustn't be NULL 
*/
/*----------------------------------------------------------------------------------------------*/
SCSCP_xmlattrptr SCSCP_xmlnode_getattr(SCSCP_xmlnodeptr curnode)
{ 
 return  curnode->properties;
}

/*----------------------------------------------------------------------------------------------*/
/*! get the content of the attribute 
  return 0 on failure and non-zero on success
  @param attr (in) current attribute. mustn't be NULL 
  @param name (out) name of this attribute
  @param value (out) value of this attribute
*/
/*----------------------------------------------------------------------------------------------*/
int SCSCP_xmlattr_getvalue(SCSCP_xmlattrptr attr, const char **name, const char **value)
{ 
 *name = xmlChar2char(attr->name);
 *value = xmlChar2char(attr->children->content);
 return  1;
}

/*----------------------------------------------------------------------------------------------*/
/*!  return the next attribute after attr 
  return NULL if attr is the last attribute
  @param attr (in) current valid node. mustn't be NULL 
*/
/*----------------------------------------------------------------------------------------------*/
SCSCP_xmlattrptr SCSCP_xmlattr_getnext(SCSCP_xmlattrptr attr)
{ 
 return attr->next;
}



/*-----------------------------------------------------------------*/
/*! free the xml document
*/
/*-----------------------------------------------------------------*/
void OML_inputstream::FreeXmlDoc()
{
 if (m_doc) ::xmlFreeDoc(m_doc);
 m_doc = NULL;
}


/*-----------------------------------------------------------------*/
/*! return an iterator on the root node of the xml document
*/
/*-----------------------------------------------------------------*/
OML_inputstream::iterator OML_inputstream::getiterator()
{
  return ::xmlDocGetRootElement(m_doc);
}

/*-----------------------------------------------------------------*/
/*! move to the child of the current object and return a new iterator
  @param iter (in) iterator
*/
/*-----------------------------------------------------------------*/
OML_inputstream::iterator OML_inputstream::GoToChild(iterator iter)
{
 return iter->children;
}

/*-----------------------------------------------------------------*/
/*!move the current node pointer to the first specified item
 return true on success
  @param iter (inout) iterator on the xml doc
  @param string (in) string to find
  @param children (in) = true => look only in the children
*/
/*-----------------------------------------------------------------*/
bool OML_inputstream::MoveToTag(iterator& iter, const_OMLchar string, bool children)
{
  bool res = false;
  iterator itertmp = iter;
  SCSCP_getdebugger()<<"OML_inputstream::MoveToTag("<<iter<<","<<string<<") - entree\n";
  if (children)
  {
      if (iter)
      {
       if (iter->type==XML_TEXT_NODE)
       { // cas de texte entre les noeuds
        res = MoveToTag(iter->next, string, children);
       }
       else if (iter->type == XML_ELEMENT_NODE) 
       {
        SCSCP_getdebugger()<<"OML_inputstream::MoveToTag() - node "<<iter->name<<" type XML_ELEMENT_NODE\n";
        res = (::xmlStrcmp((const xmlChar *)string, iter->name)==0);
       }
       if (!res) 
       {
        itertmp = iter->children;
        res = MoveToTag(itertmp, string, children);
        if (res) iter = itertmp;
       }
      }
  }
  else
  {
      while (iter!=NULL && !res)
      {
       if (iter->type == XML_ELEMENT_NODE) 
       {
        SCSCP_getdebugger()<<"OML_inputstream::MoveToTag() - node "<<iter->name<<" type XML_ELEMENT_NODE\n";
        res = (::xmlStrcmp((const xmlChar *)string, iter->name)==0);
       }
       if (!res) iter = iter->next;
      }
  }       
 SCSCP_getdebugger()<<"OML_inputstream::MoveToTag("<<iter<<",) - sortie et retourne "<<res<<"\n";
  return res;
}

 
/*-----------------------------------------------------------------*/
/*! move the current node pointer to the specified symbol
  return true on success
  @param iter (inout) iterator on the xml doc
  @param cdname (in) name of the CD
  @param name (in) name of the symbol
*/
/*-----------------------------------------------------------------*/
bool OML_inputstream::MoveToTagOMS(iterator& iter, const_OMLchar cdname, const_OMLchar name)
{
 bool res = true;
 bool bfound = false;
 SCSCP_getdebugger()<<"OML_inputstream::MoveToTagOMS("<<iter<<","<<cdname<<","<<name<<", ) - entree\n";
 while (!bfound && res)
 {
  if (MoveToTag(iter, "OMS", false))
  {
    /* check cdname and name */
    OML_OMS oms;
    if ((res=oms.Read(iter))==true)
    {
     bfound = (strcmp(cdname, oms.GetCD())==0 && strcmp(name, oms.GetName())==0);
    }
  } 
  else { break; /* exit from loop */ }
 }
 
 res = res && bfound;
 
 SCSCP_getdebugger()<<"OML_inputstream::MoveToTagOMS("<<iter<<",) - sortie et retourne "<<res<<"\n";
 return res;
}


/********************************************************/
/*!
 return the indice from the name of the planet
 @param name (in) name of the planet
*/
/********************************************************/
OML_inputstream::OMLtype OML_inputstream::GetTypeFromName(const_OMLchar name) const
{
 OMLtype indice = eOMLInvalid;
 size_t cpt;
 
 for(cpt=0; cpt<sizeof(s_tableindicename)/sizeof(s_tableindicename[0]); cpt++)
 {
  if (strcmp(s_tableindicename[cpt].name, name)==0) 
  {
   indice=s_tableindicename[cpt].type;
   break; /* exit the loop */
  }
 }
 
 /* not found => generate fatal error */
 return indice;
}

/*-----------------------------------------------------------------*/
/*! return the type of the object pointed by iter
*/
/*-----------------------------------------------------------------*/
OML_inputstream::OMLtype OML_inputstream::GetOMtype(const iterator& iter) const
{
  OMLtype etype = eOMLInvalid;
  if (iter->type == XML_ELEMENT_NODE) 
  {
   etype = GetTypeFromName(xmlChar2char(iter->name));
  }
  return etype;
}


/*-----------------------------------------------------------------*/
/*! read the specified symbol and read the associated OMI
  @param iter (inout) iterator on the xml doc
  @param cdname (in) name of the CD
  @param name (in) name of the symbol
  @param integer (out) read value
*/
/*-----------------------------------------------------------------*/
template <typename T>
bool OML_inputstream::ReadPairOMSOMItmpl(iterator& iter, const_OMLchar cdname, const_OMLchar name, T& integer)
{
 bool res;
 bool bfound;

 SCSCP_getdebugger()<<"OML_inputstream::ReadPairOMSOMItmpl("<<iter<<","<<cdname<<","<<name<<", ) - entree\n";
 bfound = MoveToTagOMS(iter, cdname, name);
 /* if found => get integer */
 if (bfound)
 {
  OML_OMI omi;
  if ((res = omi.Read(iter))==true)
  {
   res = omi.Get(integer);
  }
 }
 else res = false;
 
 SCSCP_getdebugger()<<"OML_inputstream::ReadPairOMSOMItmpl("<<iter<<","<<integer<<") - sortie et retourne "<<res<<"\n";
 return res;
}

/*-----------------------------------------------------------------*/
/*! read the specified symbol and read the associated OMI
  @param iter (inout) iterator on the xml doc
  @param cdname (in) name of the CD
  @param name (in) name of the symbol
  @param integer (out) read value
*/
/*-----------------------------------------------------------------*/
bool OML_inputstream::ReadPairOMSOMI(iterator& iter, const_OMLchar cdname, const_OMLchar name, size_t& integer)
{
 return ReadPairOMSOMItmpl(iter, cdname, name, integer);
}

/*-----------------------------------------------------------------*/
/*! read the specified symbol and read the associated OMI
  @param iter (inout) iterator on the xml doc
  @param cdname (in) name of the CD
  @param name (in) name of the symbol
  @param integer (out) read value
*/
/*-----------------------------------------------------------------*/
/*bool OML_inputstream::ReadPairOMSOMI(iterator& iter, const_OMLchar cdname, const_OMLchar name, unsigned long & integer)
{
 return ReadPairOMSOMItmpl(iter, cdname, name, integer);
}*/

/*-----------------------------------------------------------------*/
/*! read the specified symbol and read the associated OMI
  @param iter (inout) iterator on the xml doc
  @param cdname (in) name of the CD
  @param name (in) name of the symbol
  @param integer (out) read value
*/
/*-----------------------------------------------------------------*/
bool OML_inputstream::ReadPairOMSOMI(iterator& iter, const_OMLchar cdname, const_OMLchar name, int& integer)
{
 return ReadPairOMSOMItmpl(iter, cdname, name, integer);
}

/*-----------------------------------------------------------------*/
/*! read the specified symbol and read the associated OMSTR
  @param iter (inout) iterator on the xml doc
  @param cdname (in) name of the CD
  @param name (in) name of the symbol
  @param string (out) read value
*/
/*-----------------------------------------------------------------*/
bool OML_inputstream::ReadPairOMSOMSTR(iterator& iter, const_OMLchar cdname, const_OMLchar name, const_OMLchar& string)
{
 bool res = false;
 
 SCSCP_getdebugger()<<"OML_inputstream::ReadPairOMSOMSTR("<<iter<<","<<cdname<<","<<name<<", ) - entree\n";

 string = NULL;
 res = MoveToTagOMS(iter, cdname, name);
 if (res) res = OML_OMSTR::Read(iter, string);
 SCSCP_getdebugger()<<"OML_inputstream::ReadPairOMSOMSTR("<<iter<<","<<string<<") - sortie et retourne "<<res<<"\n";
 return res;
}

/*-----------------------------------------------------------------*/
/*! return the raw string assoicted to the current iterator
    the returned string must be freed by free.
  @param iter (in) iterator on the xml doc    
*/
/*-----------------------------------------------------------------*/
OML_inputstream::OMLchar OML_inputstream::GetRawString(iterator& iter)
{
 OMLchar str = NULL;
 SCSCP_getdebugger()<<" OML_inputstream::GetRawString("<<iter<<") - entree\n";
 
 xmlBufferPtr buffer = xmlBufferCreate();
 if (xmlNodeDump(buffer, m_doc, iter, 1,1)!=-1)
 {
    const xmlChar *strxml = xmlBufferContent(buffer);
    if (strxml!=NULL)
    {
     int len = xmlStrlen(strxml)+1;
     str = (OMLchar)malloc(sizeof(char)*len);
     if (str!=NULL)
     {
      strcpy(str, xmlChar2char(strxml));
     }
    }
 }
 
 SCSCP_getdebugger()<<" OML_inputstream::GetRawString("<<iter<<") - sortie et retourne '"<<str<<"'\n";
 return str;
}

/*-----------------------------------------------------------------*/
/*! return the associated value to the attributes specified by the name
  return NULL if not found
  @param iter (in) iterator on the xml doc
  @param p_pszName (in) name of the attribute
*/
/*-----------------------------------------------------------------*/
OML_inputstream::const_OMLchar OML_FindAttributes(OML_inputstream::iterator iter, OML_inputstream::const_OMLchar p_pszName)
{
 OML_inputstream::const_OMLchar attr = NULL;
 SCSCP_getdebugger()<<"OML_FindAttributes("<<iter<<","<<p_pszName<<") - entree\n";
 
  const char* szContent;
  const xmlChar* szType;
  struct _xmlAttr* pattAttribut=iter->properties;
  while (pattAttribut && attr==NULL)
  {
            szType=pattAttribut->name;
 	    szContent=xmlChar2char(pattAttribut->children->content);
            if (strcmp(xmlChar2char(szType),p_pszName)==0)
 			{
 				attr = szContent;
  			}
 			pattAttribut=pattAttribut->next;
  }

 SCSCP_getdebugger()<<"OML_FindAttributes("<<iter<<",) - sortie et retourne "<<attr<<"\n";
 return attr;
}

/*-----------------------------------------------------------------*/
/*! load one scscp message
     return false on error
     
    @param status (inout) status information 
*/
/*-----------------------------------------------------------------*/
bool OML_inputstreamparser::LoadSCSCPobject(SCSCP_status* status)
{
 bool bres;
 SCSCP_getdebugger()<<"OML_inputstreamparser::LoadSCSCPobject() - entree\n";
 bres = GetStream().ReadLine(status); //start
 if (bres) bres = GetStream().ReadXmlDoc();
 if (bres)
 {
  SCSCP_getdebugger()<<"OML_inputstreamparser::LoadSCSCPobject(() - ReadXmlDoc returns true\n";
  m_iterload = GetStream().getiterator();
  //read the omobj tag
  bres = GetStream().ReadTagOMOBJ(m_iterload);
  m_iterload = GetStream().GoToChild(m_iterload);

 }
 SCSCP_getdebugger()<<"OML_inputstreamparser::LoadSCSCPobject() - sortie et retourne "<<bres<<"\n";
 return bres;
}


