/*-----------------------------------------------------------------*/
/*! 
  \file scscpfileclient.cxx
  \brief SCSCP client over file or socket
  \author  M. Gastineau 
           Astronomie et Systemes Dynamiques, IMCCE, CNRS, Observatoire de Paris. 

   Copyright, 2008,  M. Gastineau, IMCCE-CNRS
   email of the author : gastineau@imcce.fr
*/
/*-----------------------------------------------------------------*/

/*-----------------------------------------------------------------*/
/* License  of this file :
 This library is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/
/*-----------------------------------------------------------------*/

#if HAVE_CONFIG_H
#include "scscpconfig.h"
#endif /*HAVE_CONFIG_H*/
#if HAVE_WINDOWS_H
#include <windows.h>
#endif /*HAVE_CONFIG_H*/
#include <sys/types.h>
#if HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#if HAVE_NETDB_H
#include <netdb.h>
#endif
#if HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#if HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#if HAVE_STRING_H
#include <string.h>
#endif
#include <errno.h>
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <stdarg.h>
#include "scscptypes.h"
#include "scscpinternal.h"
#include "scscpfileclient.h"
#include "scscpdebug.h"
#define __SCSCP_WITHIN_SCSCP
#include "scscp.h"

#define MAXLENBUFFER 4094

/*-----------------------------------------------------------------*/
/* private functions */
/*-----------------------------------------------------------------*/
static int SCSCP_sc_callsendstr_callback(SCSCP_io* client, void *param, SCSCP_status* status);
static int SCSCP_io_checkNULL(SCSCP_io* stream, SCSCP_status* status);

/*-----------------------------------------------------------------*/
/*! connect to a server at the specified port
    return true on success
    on error, geterror() will give the error message
    @param servername (in) name of the server
    @param port (in) port on the server of the SCSCP server
    @param status (out) status
*/
/*-----------------------------------------------------------------*/
bool SCSCP_Fileclient::Connect(const char *servername, int port, SCSCP_status* status)
{
  bool bres  = true;
  static bool binitsocket=true;
  const char *supportedversion;
  SCSCP_getdebugger()<<"SCSCP_Fileclient::connect("<<servername<<","<<port<<") - entree\n";
  
  if (binitsocket && !SCSCP_initsocket())
  {
   SCSCP_getdebugger()<<"SCSCP_Fileclient::connect() - SCSCP_initsocket failed\n";
   seterror(errno);
   SCSCP_status_seterror(status, SCSCP_STATUS_ERRNO);
   bres = false;
  }
  struct hostent * hp = gethostbyname(servername);
  if ( hp == NULL ) 
  {
   SCSCP_getdebugger()<<"SCSCP_Fileclient::connect() - gethostbyname failed\n";
   seterror(errno);
   SCSCP_status_seterror(status, SCSCP_STATUS_ERRNO);
   bres = false;
  }
  /* creer une socket pour se connecter au serveur */
  else if ((m_fd = socket(AF_INET, SOCK_STREAM, 0))==INVALID_SOCKET)
  {
   SCSCP_getdebugger()<<"SCSCP_Fileclient::connect() - socket failed\n";
   seterror(errno);
   bres = false;
  }
  else
  {
   struct sockaddr_in adresseServeur;
   adresseServeur.sin_family = AF_INET;
   adresseServeur.sin_port = htons(port);/* host to network short */
   memcpy(&adresseServeur.sin_addr.s_addr, hp->h_addr, hp->h_length);

   SCSCP_getdebugger()<<"SCSCP_Fileclient::connect() - socket sucess m_fd="<<m_fd<<"\n";
   SCSCP_getdebugger()<<"SCSCP_Fileclient::connect() - server address : "<<inet_ntoa(adresseServeur.sin_addr)<<"\n";
   /* se connecter au serveur */
   if (::connect(m_fd, (struct sockaddr *)&adresseServeur, sizeof(adresseServeur))== SOCKET_ERROR ) 
   {
    SCSCP_getdebugger()<<"SCSCP_Fileclient::connect() - connect failed\n";
    seterror(errno);
    SCSCP_status_seterror(status, SCSCP_STATUS_ERRNO);
    bres = false;
    ::closesocket(m_fd);  
    m_fd = -1; 
   }
   else
   {
    m_OutStream.SetSocket(m_fd);
    m_InStream.SetSocket(m_fd);
    bres = ReadCIM(&supportedversion, status);
    if (bres) bres = WriteCIMVersion(supportedversion);
    if (bres) bres = ReadCIMVersion(status);
    if (!bres) Close();
   }
  }
  binitsocket = false;
  SCSCP_getdebugger()<<"SCSCP_Fileclient::connect() - sortie et retourne "<<bres<<"\n";
  return bres;
}


/*-----------------------------------------------------------------*/
//! close the connection to the server
/*-----------------------------------------------------------------*/
void SCSCP_Fileclient::Close()
{
 SCSCP_getdebugger()<<"SCSCP_Fileclient::Close() - entree\n";
 if (m_fd!=-1) 
 { 
  const char *msg="<?scscp quit ?>";
  m_OutStream.GetFileio().write(msg, strlen(msg)*sizeof(char));
  ::closesocket(m_fd);  
  m_fd = -1; 
 }
 SCSCP_getdebugger()<<"SCSCP_Fileclient::Close() - sortie\n";
}

/*-----------------------------------------------------------------*/
/*! read a line from the input stream
  and return true on success
 @param buffer (out) buffer to fill
 @param lenbuffer (in) size of buffer
*/
/*-----------------------------------------------------------------*/
bool SCSCP_Fileclient::ReadLine(char *buffer, int lenbuffer)
{
    bool bres = true;
    int j=-1;
    
    SCSCP_getdebugger()<<"SCSCP_Fileclient::ReadLine() - entree\n";
    
    do
    {
     j++;
     bres=(m_InStream.GetFileio().read(buffer+j, sizeof(char))!=SOCKET_ERROR);
    } while (bres && buffer[j]!='\n' && j<lenbuffer);
    if (j==lenbuffer) bres = false;
    if (!bres)
    {
     seterror(errno);
    }
    else
    {
     buffer[j] = '\0';
     SCSCP_getdebugger()<<"SCSCP_Fileclient::ReadLine - a lu '"<<buffer<<"'\n";
    }
    
    SCSCP_getdebugger()<<"SCSCP_Fileclient::ReadLine - sortie et retourne "<<bres<<"\n";
    return bres;
}
/*-----------------------------------------------------------------*/
/*! read the Connection information Message from the server
  and return true on success
 @param supportedversion (out) supported version
  @param status (inout) status 
*/
/*-----------------------------------------------------------------*/
bool SCSCP_Fileclient::ReadCIM(const char **supportedversion,SCSCP_status* status)
{
    bool bres;
    char c[MAXLENBUFFER];
    char *scscp_versions;
    char *lasts;
    
    SCSCP_getdebugger()<<"SCSCP_Fileclient::ReadCIM() - entree\n";
 
     bres = ReadLine(c, MAXLENBUFFER);
     *supportedversion = NULL;
     if (bres)
     {
      scscp_versions = SCSCP_findversionstring(c,"scscp_versions");
      if (scscp_versions)
      {
         for (char *vers=strtok_r(scscp_versions, " ", &lasts); vers && *supportedversion==NULL; vers= strtok_r(NULL, " ", &lasts))
         {
          for (size_t j=0; j<m_allowedversionscount && *supportedversion==NULL; j++)
          {
            if (strcmp(vers,m_allowedversions[j])==0) *supportedversion = m_allowedversions[j];
          }
         }
         /* aucune version supporter => on donne la + grande */
         if (*supportedversion==NULL) *supportedversion=m_allowedversions[0];
      }
      else
      {
       bres = false;
       SCSCP_status_seterror(status, SCSCP_STATUS_VERSIONNEGOTIATIONFAILED);
      }
    }
    
    SCSCP_getdebugger()<<"SCSCP_Fileclient::ReadCIM - sortie et retourne "<<bres<<"\n";
    return bres;
}

     /* choose the version */
 /*    strcpy(c2, c);
     
*/

/*-----------------------------------------------------------------*/
/*! read the Connection information Message Version from the server
  and return true on success
  @param status (inout) status 
*/
/*-----------------------------------------------------------------*/
bool SCSCP_Fileclient::ReadCIMVersion(SCSCP_status* status)
{
    bool bres;
    char c[MAXLENBUFFER];
    char *scscp_versions;
    
    SCSCP_getdebugger()<<"SCSCP_Fileclient::ReadCIMVersion() - entree\n";
 
     bres = ReadLine(c, MAXLENBUFFER);
     if (bres)
     {
      scscp_versions = SCSCP_findversionstring(c,"version");
      if (scscp_versions==NULL) 
      {
       bres = false;
       SCSCP_status_seterror(status, SCSCP_STATUS_VERSIONNEGOTIATIONFAILED);
      }
    }
    
    SCSCP_getdebugger()<<"SCSCP_Fileclient::ReadCIMVersion - sortie et retourne "<<bres<<"\n";
    return bres;
}

/*-----------------------------------------------------------------*/
/*! write the SCSCP version of the client to the server
  and return true on success
 @param supportedversion (in) supported version
*/
/*-----------------------------------------------------------------*/
bool SCSCP_Fileclient::WriteCIMVersion(const char *supportedversion)
{
    bool bres;
    //const char *msg="<?scscp version=\"1.1\" ?>\n";
     const char *msg1="<?scscp version=\"";
     const char *msg2="\" ?>\n";
    SCSCP_getdebugger()<<"SCSCP_Fileclient::WriteCIMVersion() - entree\n";
 
    //bres = m_OutStream.GetFileio().write(msg, strlen(msg)*sizeof(char));
 
    bres =  m_OutStream.GetFileio().write(msg1, strlen(msg1)*sizeof(char));
    if (bres) bres = m_OutStream.GetFileio().write(supportedversion, strlen(supportedversion)*sizeof(char));
    if (bres) bres = m_OutStream.GetFileio().write(msg2, strlen(msg2)*sizeof(char));
     
    SCSCP_getdebugger()<<"SCSCP_Fileclient::WriteCIMVersion - sortie et retourne "<<bres<<"\n";
    return bres;
}

/*-----------------------------------------------------------------*/
/*! attach the specified to the client
    @param socket (in) socket to be attached
*/
/*-----------------------------------------------------------------*/
void SCSCP_Fileclient::SetSocket(int socket)
{
 m_fd = socket;
 m_OutStream.SetSocket(m_fd);
 m_InStream.SetSocket(m_fd);
}

/*-----------------------------------------------------------------*/
/*! checks if SCSCP client is null and returns 0 on error
  @param client (in) object to be checked
  @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_sc_checkNULL(SCSCP_socketclient* client, SCSCP_status* status)
{
   int res = 1;
   
   if (client==NULL || *client==NULL)
   {
     SCSCP_status_seterror(status,  SCSCP_STATUS_CLIENTOBJECTNULL);
     res = 0;
   }
   return res;
}

/*-----------------------------------------------------------------*/
/*! initialize the object client and set status if an error occurs 
 @param client (inout) object to be initialized
 @param status (inout) status error
 xparam ...    (in) allowed version to be negotiated
*/
/*-----------------------------------------------------------------*/
int SCSCP_sc_init(SCSCP_socketclient* client, SCSCP_status* status, ...)
{
   va_list versions;
   int res = 1;
   
   SCSCP_status_seterror(status, SCSCP_STATUS_OK);
   if (client==NULL)
   {
     SCSCP_status_seterror(status,  SCSCP_STATUS_CLIENTOBJECTNULL);
     res = 0;
   }
   else
   {
    SCSCP_Fileclient* pclient;
    *client = pclient = new SCSCP_Fileclient;
    if (pclient==NULL) 
    {
     SCSCP_status_seterror(status, SCSCP_STATUS_NOMEM);
     res = 0;
    }
    else
    {
     /* count elements */
     size_t  counter = 0;
     va_list listcount;
     va_start(listcount, status);
     while (va_arg(listcount, const char*)!=NULL) counter++;
     va_end(listcount);

     /* duplicate the elements of versions */
     pclient->m_allowedversions = (SCSCP_string*)SCSCP_malloc(sizeof(SCSCP_string)*counter, status);
     res = (pclient->m_allowedversions!=NULL);
     if (res)
     {
      va_start(versions, status);
      for(size_t cpt=0; cpt<counter; cpt++)
      {
       pclient->m_allowedversions[cpt] = SCSCP_strdup(va_arg(versions, const char*), status);
       res = (pclient->m_allowedversions[cpt]!=NULL);
      }
      pclient->m_allowedversionscount = counter;
      va_end(versions);
     }
    }
   }
   return res;
}

/*-----------------------------------------------------------------*/
/*! clear the object client previously initialized by SCSCP_sc_init  
 @param client (inout) object to be deleted
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_sc_clear(SCSCP_socketclient* client, SCSCP_status* status)
{
   SCSCP_getdebugger()<<"SCSCP_sc_clear() - entree\n";
   int res = SCSCP_sc_checkNULL(client, status);
   if (!res) return res;
   res = SCSCP_sc_clearallowedversions(*client, status);
   (*client)->Close();
   delete (*client);
   SCSCP_getdebugger()<<"SCSCP_sc_clear() - sortie\n";
   return res;
}

/*-----------------------------------------------------------------*/
/*! connect the client to a server on the specified host and port  
 @param client (inout) scscp client
 @param machine (in) computer name or IP address
 @param port (in) port on the remote computer
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_sc_connect(SCSCP_socketclient* client, const char *machine, int port, SCSCP_status* status)
{
   int res = SCSCP_sc_checkNULL(client, status);
   if (!res) return res;
   
   SCSCP_status_seterror(status, SCSCP_STATUS_OK);
   res = (*client)->Connect(machine, port,status);
   if (!res && SCSCP_status_is(status)==SCSCP_STATUS_OK) SCSCP_status_seterror(status,SCSCP_STATUS_ERRNO);
   return res;
}
 
/*-----------------------------------------------------------------*/
/*! close the connection of the client to a server 
   previously opened by SCSCP_sc_connect  
 @param client (inout) scscp client
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_sc_close(SCSCP_socketclient* client, SCSCP_status* status)
{
   int res = SCSCP_sc_checkNULL(client, status);
   if (!res) return res;
   
   (*client)->Close();
   return res;
}

/*-----------------------------------------------------------------*/
/*! write the openmath buffer to the scscp server
 @param stream (inout) scscp stream
 @param param (in) openmath expression (must be a const char *)
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
static int SCSCP_sc_callsendstr_callback(SCSCP_io* stream, void *param, SCSCP_status* status)
{
  const char *openmathbuffer = (const char*)param;
  return SCSCP_io_write(stream, openmathbuffer, status);
}




/*-----------------------------------------------------------------*/
/*! send an openmath string to the SCSCP server with some options 
    for a procedure call
 @param client (inout) scscp client
 @param options (in) options of the call
 @param openmathbuffer (in) openmath expression
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_sc_callsendstr(SCSCP_socketclient* client, SCSCP_calloptions* options, 
                          const char *openmathbuffer, SCSCP_status* status)
{
   return SCSCP_sc_callsendcallback(client, options, SCSCP_sc_callsendstr_callback, (void*)openmathbuffer, status);
}

/*-----------------------------------------------------------------*/
/*! send a procedure call to the SCSCP server with some options 
    The parameter of the procedure call are written 
    by the callback function "callbackwriteargs"
 @param client (inout) scscp client
 @param options (in) options of the call
 @param callbackwriteargs (in) callback function 
             to write the parameters of the procedure call
 @param param (inout) opaque data given to callbackwriteargs
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_sc_callsendcallback(SCSCP_socketclient* client, SCSCP_calloptions* options, 
                          int (*callbackwriteargs)(SCSCP_io* stream, void *param, SCSCP_status* status),
                          void *param, SCSCP_status* status)
{
   int res = SCSCP_sc_checkNULL(client, status);
   if (!res) return res;
   
   res = SCSCP_procedurecall_Write(*client, options==NULL?NULL:*options, callbackwriteargs, param, status);
   
   return res;
}

/*-----------------------------------------------------------------*/
/*! return the first node in the current location of the openmath message
 @param client (inout) scscp client
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
SCSCP_xmlnodeptr SCSCP_sc_getxmlnode(SCSCP_socketclient *client, SCSCP_status* status)
{
 int res = SCSCP_sc_checkNULL(client, status);
 return res?(*client)->GetIter():NULL;
}

/*-----------------------------------------------------------------*/
/*! write some data to the SCSCP server or client 
 return 0 on error
 @param client (inout) scscp stream
 @param buffer (in) some data
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_write(SCSCP_io* stream, const char *buffer, SCSCP_status* status)
{
 int res = SCSCP_io_checkNULL(stream, status);
 if (res)
 {
   OML_Fileoutputstream astream((*stream)->GetFileio(), true);
   res = astream.WriteChars(buffer);
  }
  return res;
}

/*-----------------------------------------------------------------*/
/*! checks if SCSCP stream is null and returns 0 on error
  @param client (in) object to be checked
  @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
static int SCSCP_io_checkNULL(SCSCP_io* stream, SCSCP_status* status)
{
   int res = 1;
   
   if (stream==NULL || *stream==NULL)
   {
     SCSCP_status_seterror(status,  SCSCP_STATUS_STREAMOBJECTNULL);
     res = 0;
   }
   return res;
}


