/*-----------------------------------------------------------------*/
/*! 
  \file scscpserver.cxx
  \brief SCSCP server
  \author  M. Gastineau 
           Astronomie et Systemes Dynamiques, IMCCE, CNRS, Observatoire de Paris. 

   Copyright, 2008,  M. Gastineau, IMCCE-CNRS
   email of the author : gastineau@imcce.fr
*/
/*-----------------------------------------------------------------*/

/*-----------------------------------------------------------------*/
/* License  of this file :
This file is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute this file under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/*-----------------------------------------------------------------*/

#if HAVE_CONFIG_H
#include "scscpconfig.h"
#endif /*HAVE_CONFIG_H*/
#if HAVE_WINDOWS_H
#include <windows.h>
#endif /*HAVE_CONFIG_H*/
#include <sys/types.h>
#if HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#if HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#include <errno.h>
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#if HAVE_STRING_H
#include <string.h>
#endif
#include <libxml/parser.h>
#include <libxml/tree.h>
#if HAVE_STRING_H
#include <string.h>
#endif
#include "scscptypes.h"
#include "scscpinternal.h"
#include "omlfileinputstream.h"
#include "omlfileoutputstream.h"
#include "scscpprocedurecompleted.h"
#include "scscpfileserver.h"
#include "scscpfileclient.h"
#include "scscpdebug.h"
#define __SCSCP_WITHIN_SCSCP
#include "scscp.h"

/*-----------------------------------------------------------------*/
/*! listen for the clients on the specified port
    return false on failure
    on error, geterror() will give the error message
    @param port (in) port on the server of the SCSCP server
*/
/*-----------------------------------------------------------------*/
bool SCSCP_Fileserver::Listen(int port)
{
  bool bres = true;
  static bool binitsocket=true;
  SCSCP_getdebugger()<<"SCSCP_Fileserver::Listen("<<port<<") - entree\n";

  if (binitsocket && !SCSCP_initsocket())
  {
   SCSCP_getdebugger()<<"SCSCP_Fileserver::connect() - socket failed\n";
   seterror(errno);
   bres = false;
  }
  /* creer une socket pour se connecter au serveur */
  else if ((m_fd = socket(AF_INET, SOCK_STREAM, 0))==INVALID_SOCKET)
  {
   SCSCP_getdebugger()<<"SCSCP_Fileserver::connect() - socket failed\n";
   seterror(errno);
   bres = false;
  }
  else
  {
   struct sockaddr_in adresseServeur;
   memset(&adresseServeur, 0, sizeof(adresseServeur));
   adresseServeur.sin_family = AF_INET;
   adresseServeur.sin_addr.s_addr = htonl(INADDR_ANY);
   adresseServeur.sin_port = htons(port);/* host to network short */

   SCSCP_getdebugger()<<"SCSCP_Fileserver::Listen() - socket sucess m_fd="<<m_fd<<"\n";

   /* se connecter au serveur */
   if (bind(m_fd,  (struct sockaddr *) &adresseServeur, sizeof(adresseServeur)) < 0)
   {
    SCSCP_getdebugger()<<"SCSCP_Fileserver::Listen() - bind failed\n";
    seterror(errno);
    bres = false;
   }
   else if (listen(m_fd,  5) < 0)
   {
    SCSCP_getdebugger()<<"SCSCP_Fileserver::Listen() - listen failed\n";
    seterror(errno);
    bres = false;
   }
  }
  
  binitsocket = false;
  SCSCP_getdebugger()<<"SCSCP_Fileserver::Listen() - sortie et retourne "<<bres<<"\n";
  return bres;
}


/*-----------------------------------------------------------------*/
/*! accept an incoming connection
 @param stream (inout) stream to initialize
*/
/*-----------------------------------------------------------------*/
bool SCSCP_Fileserver::Accept(OML_Fileinputstream& stream)
{
    SCSCP_getdebugger()<<"SCSCP_Fileserver::Accept() - entree\n";
    int newsocket;
    bool bres = Accept(newsocket);
    
    if (bres)
    {
     stream.SetSocket(newsocket);
    }
    
    SCSCP_getdebugger()<<"SCSCP_Fileserver::Accept() - sortie et retourne "<<bres<<"\n";
    return bres;
}


/*-----------------------------------------------------------------*/
/*! accept an incoming connection
 @param newsocket (out) socket to perform the communication
*/
/*-----------------------------------------------------------------*/
bool SCSCP_Fileserver::Accept(int& newsocket)
{
    bool bres = true;
    struct sockaddr that;
    const char *supportedversion;
    socklen_t lg = sizeof(that);
    SCSCP_getdebugger()<<"SCSCP_Fileserver::Accept() - entree\n";
    int retry=1;
    
    do
    {
		retry = 0;
		newsocket = accept(m_fd, &that, &lg);
		if (newsocket==INVALID_SOCKET)
		{
		 SCSCP_getdebugger()<<"SCSCP_Fileserver::Accept() - accept failed\n";
		 seterror(errno);
		 bres = false;
		}
		else
		{
		 bres = WriteCIM(newsocket);
		 if (bres) bres = ReadCIM(newsocket,&supportedversion);
		 if (bres) bres = WriteCIMVersion(newsocket, supportedversion);
		 if (!bres) 
		 {
			Close(newsocket);
			retry = 1;
		 }
		}
    } while(retry);
    
    SCSCP_getdebugger()<<"SCSCP_Fileserver::Accept("<<newsocket<<") - sortie et retourne "<<bres<<"\n";
    return bres;
}


/*-----------------------------------------------------------------*/
//! close the connexion to the server
/*-----------------------------------------------------------------*/
void SCSCP_Fileserver::Close()
{
 if (m_fd!=-1) { ::closesocket(m_fd);  m_fd = -1; };
}

/*-----------------------------------------------------------------*/
//! close the connexion to the client
/*-----------------------------------------------------------------*/
void SCSCP_Fileserver::Close(int socket)
{
 if (socket!=-1) 
 { 
  const char *msg="<?scscp quit ?>";
  send(socket, msg, strlen(msg)*sizeof(char), 0);
  ::closesocket(socket);  
 };
}


/*-----------------------------------------------------------------*/
/*! read the Connection information Message from the client
  choose the negotiated version
  and return true on success
 @param socket (in) socket descriptor
 @param supportedversion (out) negotiated version.
         =NULL if no common version
 @bug M. Gastineau 09/09/08 : add bres in the the while test in the case 
      that recv is failed. 
*/
/*-----------------------------------------------------------------*/
bool SCSCP_Fileserver::ReadCIM(int socket, const char** supportedversion)
{
    bool bres = true;
    char c[4094];
    char *versionstr;
    int j=-1;
    
    SCSCP_getdebugger()<<"SCSCP_Fileserver::ReadCIM("<<socket<<") - entree\n";
    do
    {
     j++;
     bres=(recv(socket, c+j, sizeof(char), 0)!=SOCKET_ERROR);
    } while (bres && c[j]!='\n' && j<4094);
    if (j==4094) bres = false;
    
    if (!bres)
    {
     seterror(errno);
    }
    else
    {
     SCSCP_getdebugger()<<"SCSCP_Fileserver::ReadCIM - a lu '"<<c<<"'\n";
     c[j] = '\0';
     versionstr = SCSCP_findversionstring(c,"version");
     *supportedversion = NULL;
     if (versionstr)
     {
      for (size_t j=0; j<m_allowedversionscount && *supportedversion==NULL; j++)
      {
       if (strcmp(versionstr,m_allowedversions[j])==0) *supportedversion = m_allowedversions[j];
      }
     }
    }
    
    SCSCP_getdebugger()<<"SCSCP_Fileserver::ReadCIM(,"<<supportedversion<<") - sortie et retourne "<<bres<<"\n";
    return bres;
}


/*-----------------------------------------------------------------*/
/*! write the Connection information Message to the client
  and return true on success
 @param socket (in) socket descriptor
*/
/*-----------------------------------------------------------------*/
bool SCSCP_Fileserver::WriteCIM(int socket)
{
    bool bres;
   // const char *msg="<?scscp service_name=\"xx\" service_version=\"0\" service_id=\"0\" scscp_versions=\"1.1\" ?>\n";
     const char *msg1="<?scscp service_name=\"";
     const char *msg2="\" service_version=\"";
     const char *msg3="\" service_id=\"";
     const char *msg4="\" scscp_versions=\"";
     const char *msg5="\" ?>\n";
     SCSCP_getdebugger()<<"SCSCP_Fileserver::WriteCIM("<<socket<<") - entree\n";
 
    //bres = (send(socket, msg, strlen(msg)*sizeof(char), 0)!=SOCKET_ERROR);
    bres = (send(socket, msg1, strlen(msg1)*sizeof(char), 0)!=SOCKET_ERROR);
    if (bres) bres = (send(socket, m_servicename, strlen(m_servicename)*sizeof(char), 0)!=SOCKET_ERROR);
    
    if (bres) bres = (send(socket, msg2, strlen(msg2)*sizeof(char), 0)!=SOCKET_ERROR);
    if (bres) bres = (send(socket, m_serviceversion, strlen(m_serviceversion)*sizeof(char), 0)!=SOCKET_ERROR);
    
    if (bres) bres = (send(socket, msg3, strlen(msg3)*sizeof(char), 0)!=SOCKET_ERROR);
    if (bres) bres = (send(socket, m_serviceid, strlen(m_serviceid)*sizeof(char), 0)!=SOCKET_ERROR);
    
    if (bres) bres = (send(socket, msg4, strlen(msg4)*sizeof(char), 0)!=SOCKET_ERROR);
    for (size_t j=0; j<m_allowedversionscount && bres; j++)
    {
      bres = (send(socket, m_allowedversions[j], strlen(m_allowedversions[j])*sizeof(char), 0)!=SOCKET_ERROR);
      if (bres) bres = (send(socket, " ", strlen(" ")*sizeof(char), 0)!=SOCKET_ERROR);
    }
    
    if (bres) bres = (send(socket, msg5, strlen(msg5)*sizeof(char), 0)!=SOCKET_ERROR);
    
    SCSCP_getdebugger()<<"SCSCP_Fileserver::WriteCIM - sortie et retourne "<<bres<<"\n";
    return bres;
}

/*-----------------------------------------------------------------*/
/*! write the SCSCP version of the server the client
  and return true on success
  if (supportedversion==NULL) => close the session with "scscp quit" 
  and return false
 @param socket (in) socket descriptor
 @param supportedversion (in) negotiated version.
         =NULL if no common version => close connection
*/
/*-----------------------------------------------------------------*/
bool SCSCP_Fileserver::WriteCIMVersion(int socket, const char *supportedversion)
{
    bool bres;
    const char *msg1="<?scscp version=\"";
     const char *msg2="\" ?>\n";
    const char *msg3="<?scscp quit reason=\"unsupported version\" ?>\n";
    SCSCP_getdebugger()<<"SCSCP_Fileserver::WriteCIMVersion("<<socket<<","<<supportedversion<<") - entree\n";
 
    if (supportedversion!=NULL)
    {
     bres = (send(socket, msg1, strlen(msg1)*sizeof(char), 0)!=SOCKET_ERROR);
     if (bres) bres = (send(socket, supportedversion, strlen(supportedversion)*sizeof(char), 0)!=SOCKET_ERROR);
     if (bres) bres = (send(socket, msg2, strlen(msg2)*sizeof(char), 0)!=SOCKET_ERROR);
    }
    else
    {
     bres = (send(socket, msg3, strlen(msg3)*sizeof(char), 0)!=SOCKET_ERROR);    
     bres = false; /* => force close connection */
    }
    SCSCP_getdebugger()<<"SCSCP_Fileserver::WriteCIMVersion - sortie et retourne "<<bres<<"\n";
    return bres;
}

/*-----------------------------------------------------------------*/
/*! checks if SCSCP server is null and returns 0 on error
  @param server (in) object to be checked
  @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_ss_checkNULL(SCSCP_socketserver* server, SCSCP_status* status)
{
   int res = 1;
   
   if (server==NULL || *server==NULL)
   {
     SCSCP_status_seterror(status,  SCSCP_STATUS_SERVEROBJECTNULL);
     res = 0;
   }
   return res;
}


/*-----------------------------------------------------------------*/
/*! initialize the object server and set status if an error occurs 
 @param server (inout) object to be initialized
 @param servicename (in) name of service
 @param serviceversion (in) service version
 @param serviceid (in) service version
 @param status (inout) status error
 @param ...    (in) allowed version to be negotiated
*/
/*-----------------------------------------------------------------*/
int SCSCP_ss_init(SCSCP_socketserver* server, SCSCP_status* status, 
                  const char *servicename, const char *serviceversion, const char *serviceid, ...)
{
   va_list versions;
   int res = 1;
   
   SCSCP_status_seterror(status, SCSCP_STATUS_OK);
   if (server==NULL)
   {
     SCSCP_status_seterror(status,  SCSCP_STATUS_SERVEROBJECTNULL);
     res = 0;
   }
   else
   {
    SCSCP_socketserver pserver;
    *server = pserver = new SCSCP_Fileserver;
    if (pserver==NULL) 
    {
     SCSCP_status_seterror(status, SCSCP_STATUS_NOMEM);
     res = 0;
    }
    else
    {
     /* count elements */
     size_t  counter = 0;
     va_list listcount;
     va_start(listcount, serviceid);
     while (va_arg(listcount, const char*)!=NULL) counter++;
     va_end(listcount);

     /* duplicate the elements of versions */
     pserver->m_allowedversions = (SCSCP_string*)SCSCP_malloc(sizeof(SCSCP_string)*counter, status);
     res = (pserver->m_allowedversions!=NULL);
     if (res)
     {
      va_start(versions, serviceid);
      for(size_t cpt=0; cpt<counter; cpt++)
      {
       pserver->m_allowedversions[cpt] = SCSCP_strdup(va_arg(versions, const char*), status);
       res = (pserver->m_allowedversions[cpt]!=NULL);
     }
      pserver->m_allowedversionscount = counter;
      va_end(versions);
     }
     /* duplicate information about the services */
     if (res)
     {
       SCSCP_ss_setservice(pserver,servicename, serviceversion, serviceid, status);
     }
    }
   }
   return res;
}

/*-----------------------------------------------------------------*/
/*! clear the object server previously initialized by SCSCP_ss_init  
 @param server (inout) object to be deleted
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_ss_clear(SCSCP_socketserver* server, SCSCP_status* status)
{
   SCSCP_socketserver pserver;
   int res = SCSCP_ss_checkNULL(server, status);
   if (!res) return res;

   pserver = *server;
   SCSCP_status_seterror(status, SCSCP_STATUS_OK);
   if (pserver->m_servicename) free(pserver->m_servicename);
   if (pserver->m_serviceversion) free(pserver->m_serviceversion);
   if (pserver->m_serviceid) free(pserver->m_serviceid);
   res = SCSCP_ss_clearallowedversions(pserver, status);
   delete pserver;

   return res;
}

/*-----------------------------------------------------------------*/
/*! the object server listen on the specified port  
 @param server (inout) server 
 @param port (in) port number
 @param firstavailable (in) = 0 => if the port isn't available, return an error
                            = 1 => retry until a port is available (such that >port)
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_ss_listen(SCSCP_socketserver* server, int port, int firstavailable, SCSCP_status* status)
{
   SCSCP_socketserver pserver;
   int res = SCSCP_ss_checkNULL(server, status);
   int curport = port;
   if (res) 
   {
      pserver = *server;
      do
      { 
       res = pserver->Listen(curport);
       if (res==0 && firstavailable==1) curport++;
      } while (res==0 && firstavailable==1 && curport>port);
      //if (res) pserver->SetPort(port);
   }
   return res;
}

/*-----------------------------------------------------------------*/
/*! the object server terminates to listen  
 @param server (inout) server
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_ss_close(SCSCP_socketserver* server, SCSCP_status* status)
{
   SCSCP_socketserver pserver;
   int res = SCSCP_ss_checkNULL(server, status);
   if (res) 
   {
      pserver = *server;
      pserver->Close();
   }
   return res;
}

/*-----------------------------------------------------------------*/
/*! the object waits for an incoming connection 
   and accepts a incoming client connection  
 @param server (inout) server
 @param incomingclient (out) new client
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_ss_acceptclient(SCSCP_socketserver* server, SCSCP_incomingclient* incomingclient, SCSCP_status* status)
{
   SCSCP_socketserver pserver;
   int res = SCSCP_ss_checkNULL(server, status);
   if (res) 
   {
      pserver = *server;
      int newsocket;
      res = pserver->Accept(newsocket);
      if (res)
      {
       *incomingclient = new SCSCP_Fileclient;
       if (*incomingclient!=NULL)
       {
        (*incomingclient)->SetSocket(newsocket);
       }
       else
       {
        SCSCP_status_seterror(status, SCSCP_STATUS_NOMEM);
        res = 0;
       }
      }
   }
   return res;
}


/*-----------------------------------------------------------------*/
/*! receive the "procedure call" message from the client :
  read the attribute of the incoming procedure call and the type of the incoming message 
  store the arguments in the openmathbuffer 
 @param client (inout) scscp client
 @param options (inout) options of the procedure return
 @param msgtype (out) type of returned message
 @param openmathbuffer (out) openmath buffer. 
                 This buffer must be freed with 'free'
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_ss_callrecvstr(SCSCP_incomingclient* client, SCSCP_returnoptions* options,
                         SCSCP_msgtype *msgtype, char **openmathbuffer, SCSCP_status* status)
{
 int res = SCSCP_sc_callrecvheader(client, options, msgtype, status);
 if (res)
 {
  SCSCP_socketclient pclient = *client;
  *openmathbuffer = pclient->GetInputStream()->GetRawString(pclient->GetIter());
 }
 return res;
}

/*-----------------------------------------------------------------*/
/*!  receive the "procedure call" message from the client : 
    read the attribute of the incoming procedure call and the type of the incoming message 
  @param client (inout) scscp client
  @param options (inout) options of the procedure return
  @param msgtype (out) type of returned message.
   it must always be SCSCP_msgtype_ProcedureCall in the current specification
  @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_ss_callrecvheader(SCSCP_incomingclient* client, SCSCP_calloptions* options,
                            SCSCP_msgtype *msgtype, SCSCP_status* status)
{
   SCSCP_incomingclient pclient;
   int res;
   
   SCSCP_debugprint("SCSCP_ss_callrecvheader(%d) - start\n");
   SCSCP_status_seterror(status, SCSCP_STATUS_OK);
   res = SCSCP_sc_checkNULL(client, status);
   
   if (res) 
   {
    pclient = *client;
    SCSCP_procedurecalloptions localoptions;
    SCSCP_procedurearguments   m_ProcedureArguments;
    SCSCP_procedurecall acall(localoptions, m_ProcedureArguments);
    res = acall.ReadHeader(*(pclient->GetInputStream()), pclient->GetIter(), status);
    if (res && options!=SCSCP_CALLOPTIONS_IGNORE) **options = localoptions;
    *msgtype = SCSCP_msgtype_ProcedureCall;
   }
   
   SCSCP_debugprint("SCSCP_ss_callrecvheader(%d) - returns %d\n",SCSCP_status_is(status), res);
   return res;
}

/*-----------------------------------------------------------------*/
/*! write the openmath buffer to the scscp client
 @param client (inout) scscp incoming client
 @param param (in) openmath expression (must be a const char *)
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
static int SCSCP_ss_callsendstr_callback(SCSCP_incomingclient* client, void *param, SCSCP_status* status)
{
  const char *openmathbuffer = (const char*)param;
  return SCSCP_io_write(client, openmathbuffer, status);
}

/*-----------------------------------------------------------------*/
/*!the server send the "procedure terminated" message
    with an error type and a message to the client 
 @param client (inout) scscp client
 @param options (in) options of the call
 @param errortype (in) error code scscp2
 @param openmathbuffer (in) openmath expression
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_ss_sendterminatedstr(SCSCP_incomingclient* client, SCSCP_returnoptions* options,  
                                SCSCP_error errortype,
                                const char *messagebuffer, SCSCP_status* status)
{
   SCSCP_incomingclient pclient;
   int res = SCSCP_sc_checkNULL(client, status);
   
   if (res) 
   {
    pclient = *client;
    SCSCP_procedureterminatedoptions localoptions;
    if (options!=SCSCP_RETURNOPTIONS_DEFAULT) localoptions = **options;
    
    SCSCP_procedureterminated aprocterm(localoptions, errortype,messagebuffer);
    res = aprocterm.Write(*(pclient->GetOutputStream()));
   }
   
   return res;
}

/*-----------------------------------------------------------------*/
/*! the server send the "procedure completed" message  
    with a openmath expression as arguments  to the client
 @param client (inout) scscp client
 @param options (in) options of the call
 @param openmathbuffer (in) openmath expression
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_ss_sendcompletedstr(SCSCP_incomingclient* client, SCSCP_returnoptions* options, 
                              const char *openmathbuffer, SCSCP_status* status)
{
   return SCSCP_ss_sendcompletedcallback(client, options, SCSCP_ss_callsendstr_callback, (void*)openmathbuffer, status);
}


/*-----------------------------------------------------------------*/
/*! the server send the "procedure completed" message to the client. 
    The arguments  are written by callbackwriteargs
 @param client (inout) scscp client
 @param options (in) options of the call
 @param callbackwriteargs (in) callback function 
             to write the parameters of the procedure call
 @param param (inout) opaque data given to callbackwriteargs
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_ss_sendcompletedcallback(SCSCP_incomingclient* client, SCSCP_returnoptions* options, 
                          int (*callbackwriteargs)(SCSCP_io* stream, void *param, SCSCP_status* status),
                          void *param, SCSCP_status* status)
{
   int res = SCSCP_sc_checkNULL(client, status);
   if (!res) return res;
   
   res = SCSCP_procedurecompleted_Write(*client, options==NULL?NULL:*options, callbackwriteargs, param, status);
   
   return res;
}

/*-----------------------------------------------------------------*/
/*! close an incoming connection 
 @param client (inout) scscp client
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_ss_closeincoming(SCSCP_incomingclient* client, SCSCP_status* status)
{
 int res = SCSCP_sc_checkNULL(client, status);
 if (res)
 {
   SCSCP_sc_close(client, status);
   res = SCSCP_sc_clear(client, status);
 }
 return res;
}

/*-----------------------------------------------------------------*/
/*! return the first node in the current location of the openmath message
 @param client (inout) incoming client
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
SCSCP_xmlnodeptr SCSCP_ss_getxmlnode(SCSCP_incomingclient *client, SCSCP_status* status)
{
 int res = SCSCP_sc_checkNULL(client, status);
 return res?(*client)->GetIter():NULL;
}


