/*-----------------------------------------------------------------*/
/*! 
  \file scscptags.cxx
  \brief SCSCP tags
  \author  M. Gastineau 
           Astronomie et Systemes Dynamiques, IMCCE, CNRS, Observatoire de Paris. 

   Copyright, 2008,  M. Gastineau, IMCCE-CNRS
   email of the author : gastineau@imcce.fr
  
  \bug M. GASTINEAU 12/09/08 : update for SCSCP 1.2
*/
/*-----------------------------------------------------------------*/

/*-----------------------------------------------------------------*/
/* License  of this file :
 This library is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/
/*-----------------------------------------------------------------*/

#include <stdio.h>
#if HAVE_CONFIG_H
#include "scscpconfig.h"
#endif /*HAVE_CONFIG_H*/
#if HAVE_WINDOWS_H
#include <windows.h>
#include <Winsock2.h>
#else
#if HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#endif /*HAVE_WINDOWS_H*/
#if HAVE_STRING_H
#include <string.h>
#endif
#include <ctype.h>
#include "scscptypes.h"
#include "scscpdebug.h"
#include "scscpinternal.h"
#include "omlinputstream.h"
#define __SCSCP_WITHIN_SCSCP
#include "scscp.h"
#include "scscptags.h"


/*-----------------------------------------------------------------*/
/* variables globales*/
/*-----------------------------------------------------------------*/

/*-----------------------------------------------------------------*/
/*CD cascall1 */
/*-----------------------------------------------------------------*/
const char* SCSCP_tags::cdname = "scscp1";
const char* SCSCP_tags::call_ID = "call_ID";
const char* SCSCP_tags::option_runtime = "option_runtime";
const char *SCSCP_tags::option_min_memory = "option_min_memory";
const char *SCSCP_tags::option_max_memory = "option_max_memory";
const char *SCSCP_tags::option_return_object = "option_return_object";
const char *SCSCP_tags::option_return_cookie = "option_return_cookie";
const char *SCSCP_tags::option_return_nothing = "option_return_nothing";
const char *SCSCP_tags::option_debuglevel = "option_debuglevel";
const char *SCSCP_tags::procedure_call = "procedure_call";
const char *SCSCP_tags::procedure_completed = "procedure_completed";
const char *SCSCP_tags::procedure_terminated = "procedure_terminated"; 
const char *SCSCP_tags::info_runtime = "info_runtime";
const char *SCSCP_tags::info_memory = "info_memory";
const char *SCSCP_tags::error_memory = "error_memory";
const char *SCSCP_tags::error_runtime = "error_runtime";
const char *SCSCP_tags::error_system_specific = "error_system_specific";


/*-----------------------------------------------------------------*/
/*! initialize the windows socket DLL
    return true on success
*/
/*-----------------------------------------------------------------*/
bool SCSCP_initsocket()
{
  SCSCP_getdebugger()<<"SCSCP_initsocket() - entree\n";
#if HAVE_WINDOWS_H
 WORD wVersionRequested;
 WSADATA wsaData;
 int err;
 
 wVersionRequested = MAKEWORD( 2, 2 );
 
  err = WSAStartup( wVersionRequested, &wsaData );
  if ( err != 0 ) {
    /* Tell the user that we could not find a usable */
    /* WinSock DLL.                                  */
    SCSCP_getdebugger()<<"SCSCP_initsocket() - sortie et retourne false (WSAStartup failed) \n";
    return false;
  }
 
/* Confirm that the WinSock DLL supports 2.2.*/
/* Note that if the DLL supports versions greater    */
/* than 2.2 in addition to 2.2, it will still return */
/* 2.2 in wVersion since that is the version we      */
/* requested.                                        */
 
 if ( LOBYTE( wsaData.wVersion ) != 2 ||HIBYTE( wsaData.wVersion ) != 2 ) 
 {
    /* Tell the user that we could not find a usable */
    /* WinSock DLL.                                  */
    WSACleanup( );
    SCSCP_getdebugger()<<"SCSCP_initsocket() - sortie et retourne false (version !=2.2) \n";
    return false; 
 }

#endif
  SCSCP_getdebugger()<<"SCSCP_initsocket() - sortie et retourne true\n";
 return true;
}

/*-----------------------------------------------------------------*/
/*!  allocate memory of ncount bytes (similar to malloc) 
  @param ncount (in) size in bytes to allocate
  @param status (inout) status
*/
/*-----------------------------------------------------------------*/
void* SCSCP_malloc(size_t ncount, SCSCP_status* status)
{
 void *pres = malloc(ncount);
 if (pres==NULL) SCSCP_status_seterror(status, SCSCP_STATUS_NOMEM);
 return pres;
}
 
/*-----------------------------------------------------------------*/
/*!  duplicate a string (similar to strdup) 
  @param src (in) string to duplicate
  @param status (inout) status
*/
/*-----------------------------------------------------------------*/
char* SCSCP_strdup(const char* src, SCSCP_status* status)
{
  char *dst= (char*)SCSCP_malloc(strlen(src), status);
  if (dst) strcpy(dst, src);
  return dst; 
}

/*-----------------------------------------------------------------*/
/*!  set the error to value
  @param status (in) status to be updated
  @param value (in) error value (must be SCSCP_STATUS_xxxx)
*/
/*-----------------------------------------------------------------*/
void SCSCP_status_seterror(SCSCP_status* status, int value)
{
 if (status) *status = value;
}

/*-----------------------------------------------------------------*/
/*!  find the beginning of the version string in c.
  replace the last " of the versions by '\0'.
  return the address just after version="
  return NULL on error.
  @param c (in) status to be updated
  @param scscp_versions (in) name of string version
*/
/*-----------------------------------------------------------------*/
char * SCSCP_findversionstring(char *c, const char*versionname)
{
  char *psubstr = NULL;
  char *lastquote;
  const char *begin="<?scscp";
  
  SCSCP_getdebugger()<<"SCSCP_findversionstring("<<c<<","<<versionname<<") - entree\n";
  
  if (strncmp(begin, c,strlen(begin))==0) psubstr = c+strlen(begin);
  
  if (psubstr)
  {
      while (*psubstr!='\0' && isspace(*psubstr)) psubstr++;
    
      while (psubstr!=NULL && psubstr!='\0' 
          && strncmp(versionname, psubstr,strlen(versionname))!=0)
      {
        psubstr=strchr(psubstr,'\"');
        if (psubstr==NULL) break;
        psubstr=strchr(psubstr+1,'\"');
        if (psubstr==NULL) break;
        psubstr++;
        while (*psubstr!='\0' && isspace(*psubstr)) psubstr++;
      }
      
      
      if (psubstr!=NULL && *psubstr!='\0')
      {
       psubstr=strchr(psubstr,'\"');
       if (psubstr!=NULL)
        {
         psubstr++;
         lastquote = strchr(psubstr,'\"');
         if (lastquote) 
           *lastquote = '\0';
         else 
           psubstr = NULL;
        }
      }
      else
      {
       psubstr = NULL;
      }  
  }
  
  SCSCP_getdebugger()<<"SCSCP_findversionstring() - sortie et retourne "<<psubstr<<"\n";
  return psubstr;
}


